'use strict';

const _ = require('lodash');

const Config = require('./config');
const Microkit = require('./microkit');

/**
 * A Configurator is used to configure microkit
 */
class Configurator {
  constructor(factory) {
    this._factory = factory;
    this._options = {};
  }

  /**
   * Includes configuration from external module
   */
  include(module) {
    const optionsModule = require(module);

    let extraOptions;
    if (_.isFunction(optionsModule)) {
      extraOptions = optionsModule(this.options);
    } else {
      extraOptions = optionsModule;
    }

    this.merge(extraOptions);
  }

  /**
   * Loads configuration from config
   */
  load(options) {
    const config = new Config(options);

    this.merge(config.get('microkit'));
  }

  /**
   * Merges options with already defined options
   */
  merge(options) {
    _.defaultsDeep(_.cloneDeep(options), this._options);
  }

  validate() {
    Joi.assert(this.options, Joi.object({
      providers: Joi.object().pattern(/.+/, Joi.object({
        type: Joi.string().required()
      }).pattern(/.*/, Joi.any())).default({}),
      controllers: Joi.object().pattern(/.+/, Joi.object({
        type: Joi.string().required()
      }).pattern(/.*/, Joi.any())).default({})
    }));
  }

  /**
   * Configures microkit instance
   * @param {Microkit} microkit microkit to configure
   * @return {Microkit} configured microkit
   */
  configure(microkit) {
    _.forEach(this.options.providers, (options, name) => {
      microkit.createProvider(name, options.type, options);
    });

    _.forEach(this.options.controllers, (options, name) => {
      microkit.createController(name, options.type, options);
    });
  }
}

module.exports = Configurator;
