'use strict';

const Joi = require('joi');
const _ = require('lodash');

const microkit = require('../../');

class ExpressController extends microkit.Controller {
  initialize() {
    _.forEach(this.options.events, event => {
      this.installHandler(event, this.options.restartOnException);
    });
  }

  get schema() {
    return Joi.object({
      defaultLogFields: Joi.array().items(Joi.string()).default(
        ['method', 'host', 'protocol', 'url', 'ip']
      ),
      logFields: Joi.array().items().default([])
    });
  }

  get preHandler() {
    return (req, res, next) => {
      // catch async express errors
      const reqDomain = domain.create();
      reqDomain.on('error', next);

      // set real ip
      req.connection.ip = req.headers['x-forwarded-for'] || req.connection.remoteAddress;

      const http = _.cloneDeep(this._filterFields(
        this._parseRequest(req),
        this.options.logFields.concat(this.options.defaultLogFields)
      ));

      // attach logger to request
      req.microkit = this.microkit.extend({
        labels: {
          name: 'request',
          id: req.headers['x-request-id']
        }
      });
      req.logger = req.microkit.logger;
      req.error = req.microkit.error;

      let start;
      onFinished(res, (err, res) => {
        const end = new Date() - start;
        req.logger.info(
          `${http.method} ${http.url} ${res.statusCode} ${end}ms`,
          _.extend(http, {time: end, status: res.statusCode})
        );
      });

      req.logger.debug(`${http.method} ${http.url}`, http);
      start = new Date();
      return reqDomain.run(next);
    };
  }

  get postHandler() {
    return (err, req, res, next) => {
      const status = err.status || err.statusCode || err.status_code || 500;

      // skip anything not marked as an internal server error
      if (status < 500) {
        return next(err);
      }

      const info = this._parseRequest(req);

      req.error.capture(err, info).finally(() => next(err, req, res));
    };
  }

  _filterFields(obj, fields) {
    const result = {};
    _.forEach(fields, field => {
      _.set(result, field, _.get(obj, field));
    });
    return result;
  }

  _parseRequest(req) {
    var http = {};

    // headers:
    //
    //   node: req.headers
    //   express: req.headers
    //   koa: req.header
    //
    http.headers = req.headers || req.header || {};

    // method:
    //
    //   node: req.method
    //   express: req.method
    //   koa: req.method
    //
    http.method = req.method;

    // host:
    //
    //   node: req.headers.host
    //   express: req.hostname in > 4 and req.host in < 4
    //   koa: req.host
    //
    http.host = req.hostname || req.host || http.headers.host || '<no host>';

    // protocol:
    //
    //   node: <n/a>
    //   express: req.protocol
    //   koa: req.protocol
    //
    http.protocol = 'https' === req.protocol || true === req.secure || true === (req.socket || {}).encrypted ? 'https' : 'http';

    // url (including path and query string):
    //
    //   node: req.originalUrl
    //   express: req.originalUrl
    //   koa: req.url
    //
    http.originalUrl = req.originalUrl || req.url;

    // absolute url
    http.url = http.protocol + '://' + http.host + http.originalUrl;

    // query string
    //
    //   node: req.url (raw)
    //   express: req.query
    //   koa: req.query
    //
    http.query = req.query || urlParser.parse(req.originalUrl || '', true).query;

    // cookies:
    //
    //   node: req.headers.cookie
    //   express: req.headers.cookie
    //   koa: req.headers.cookie
    //
    http.cookies = cookie.parse(req.headers.cookie || '');

    // body data:
    //
    //   node: req.body
    //   express: req.body
    //   koa: req.body
    //
    http.data = req.body;
    if (['GET', 'HEAD'].indexOf(req.method) === -1) {
      if (typeof http.data === 'undefined') {
        http.data = '<unavailable>';
      }
    }

    if (http.data && {}.toString.call(http.data) !== '[object String]') {
      // Make sure the request body is a string
      http.data = JSON.stringify(http.data);
    }

    // client ip:
    //
    //   node: req.connection.remoteAddress
    //   express: req.ip
    //   koa: req.ip
    //
    http.ip = req.ip || (req.connection || {}).remoteAddress;

    return http;
  }
}

module.exports = ExpressController;
