'use strict';

const Promise = require('bluebird');

const Controller = require('../controller');

/**
 * Logging controller
 */
class LogController extends Controller {
  get kind() {
    return 'log';
  }

  get schema() {
    return Joi.object({
      level: Joi
        .string().default('info')
        .description('logging level'),
      context: Joi
        .object().default({})
        .description('additional context to pass to log messages'),
      levels: Joi
        .object().pattern(/w+/, Joi.number().integer().required())
        .default({
          trace: 0,
          error: 1,
          warn: 2,
          info: 3,
          debug: 4
        })
        .description('log level mappings from names to levels'),
      labels: Joi
        .object().pattern(/.+/, Joi.string())
        .description('additional labels to apply to log messages')
    });
  }

  _levelToInteger(level) {
    return this.levels[level];
  }

  _getInfo() {
    return _.extend({
      labels: this.options.labels,
      pid: process.pid,
      memoryUsage: process.memoryUsage(),
      uptime: process.uptime(),
      time: new Date().getTime(),
      host: os.hostname(),
      service: this.microkit.name,
    }, this.options.defaultInfo);
  }

  _serializeError(error) {
    return _.has(error, 'toJSON') ? error.toJSON() : error;
  }

  _parseArgs(args) {
    args = Array.apply(null, args);
    if (_.isObject(args[0])) {
      return args[0];
    }

    const msgArgs = _.takeWhile(args, arg => !_.isObject(arg));
    const context = _.takeWhile(
      args.slice(msgArgs.length),
      arg => _.isObject(arg) && !_.isError(arg)
    );
    const error = this._serializeError(
      _.find(
        args.slice(msgArgs.length),
        arg => _.isError(arg)
      )
    );

    return {
      error: error,
      msg: util.format.apply(null, msgArgs),
      context: _.extend.apply(null, context)
    };
  }

  _log(level, args) {
    const parsedArgs = this._parseArgs(args);
    const logNumberLevel = this.options.levels[level];
    const configNumberLevel = this.options.levels[this.options.level];

    // log only if message log level is equal or bigger as config log level
    if (configNumberLevel >= logNumberLevel) {
      const info = _.extend(this._getInfo(), {
        level: level,
        numericLevel: logNumberLevel
      });

      return Promise.map(this.providers, provider => provider.log(
        info,
        parsedArgs.msg,
        _.extend(parsedArgs.context, this.options.context),
        parsedArgs.error
      ));
    }
  }

  debug() {
    return this._log('debug', arguments);
  }

  info() {
    return this._log('info', arguments);
  }

  warn() {
    return this._log('warn', arguments);
  }

  error() {
    return this._log('error', arguments);
  }
}

module.exports = LogController;
