'use strict';

const Controller = require('./controller');
const ApiController = require('./controllers/api');

/**
 * Controller for event providers
 */
class EventsController extends Controller {
  intialize() {
    if(this.options.api) {
      this.api = this.microkit.getController(this.options.api);

      Joi.assert(this.api, Joi.object().type(ApiController).required(), 'no ApiController provided');
    }
  }

  _getMapping(event, scope) {
    scope = scope || 'default';

    return _.find(this.options.mappings, mapping => {
      const mappingScope = mapping.scope || 'default';
      return _.every(mapping.selector, event) && scope === mappingScope;
    });
  }

  get schema() {
    return super.schema.keys({
      api: Joi.string().description('name of the api controller to use for fething additional data'),
      mappings: Joi.array().items(Joi.object({
        selector: Joi.object().required().description('event selector for which mapping applies to'),
        scope: Joi.string().required().description('name of the scope'),
        service: Joi.string().required().description('name of the service to call'),
        method: Joi.string().required().description('method to call on service'),
        params: Joi.string().optional().description('json schema parameters')
      })).description('mappings from events to api calls').default([])
    });
  }

  /**
   * Subscribe for events
   *
   * @param {Object} selector event selector
   * @param {function} handler event handler
   * @param {Object} options additional subscription options
   */
  subscribe(selector, handler, options) {
    return this.provider.subscribe(selector, handler, options);
  }

  /**
   * Publishes event
   *
   * @param {Event} event event to publish
   * @param {Object} data data associated with event
   *
   * @return {Promise} publish information
   */
  publish(event, data) {
    Joi.assert(event, Joi.object().required(), 'event must be object');
    Joi.assert(data, Joi.object().optional(), 'event data must be object');

    return this.provider.publish(event, data);
  }

  /**
   * Fetch data associated with event
   *
   * @param {Event} event event object
   * @param {string} scope name of the event scope to fetch
   */
  fetch(event, scope) {
    if (!this.api) {
      throw new Error('Fetching not possible, ApiController not provided');
    }

    const mapping = this._getMapping(event, scope);

    if (!mapping) {
      throw new Error('Mapping not found');
    }

    const params = templateJSON(mapping.params, event);
    return this.api.call(mapping.service, mapping.method, params);
  }
}

module.exports = EventsController;
